#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
// Windows Header Files:
#include <windows.h>
#include <string>
#include "ModifierApi.h"

using namespace std;

//Modifier Information
static const char	* 	g_szName			= "SampleModifier";
static const char	*	g_szFriendlyName	= "Sample Modifier";
static const char	* 	g_szAuthor			= "Lestat";
static const char	* 	g_szDescription		= "A sample modifier";
static HINSTANCE		g_hDllInstance		= NULL;

//GetName must return the name of your modifier, only character and number are allowed, no space or accent
EXPORT_PSM const char 	* 	GetName()			{ return g_szName; }
//GetFriendly name must return the Friendly name of your modifier, this one can contain whatever you want
//it's the one that the user will see
EXPORT_PSM const char	*	GetFriendlyName()	{ return g_szFriendlyName; }
//GetAuthor must return the author name
EXPORT_PSM const char 	* 	GetAuthor()			{ return g_szAuthor; }
//GetDescription must return the description of the modifier that will be display in the modifier setup
EXPORT_PSM const char 	* 	GetDescription()	{ return g_szDescription; }
//GetModifierVersion must return version information created with MakePsmVersion
EXPORT_PSM PSMVERSION		GetModifierVersion(){ return MakePsmVersion(1,0); }
//GetApiVersion must always return MODIFIER_API_VERSION defined in ModifierApi.h
EXPORT_PSM PSMVERSION		GetApiVersion()		{ return MODIFIER_API_VERSION; }

EXPORT_PSM void OnInstall()
{
	/*
	OnInstall() is called by pogoshell studio when the user add your modifier in the settings,
	it could be a good place to initialise the default value in the config file if you need it.
	If you have nothing to do, just delete OnInstall()
	*/
}

EXPORT_PSM void OnUnInstall()
{
	/*
	OnUnInstall() is called by pogoshell studio when the user remove your modifier in the settings,
	it could be a good place to the config file if you need it.
	If you have nothing to do, just delete OnUnInstall()
	*/
}

EXPORT_PSM bool ConfigureModifier(HWND hParent)
{
	/*
	This function is called when the user click on configure in the modifier settings, if you don't 
	implement this function, then the button will be disable.
	Usually this configuration dialog is used to set default option for new modifier instance or that 
	are global to the modifier like the default path to find ips patch for example
	*/

	//DialogBoxParam(g_hDllInstance,MAKEINTRESOURCE(IDD_DIALOG_CONFIGURE_MODIFIER),hParent,(DLGPROC)ConfigureInstanceDlgProc,(LPARAM)hInstance);

	return true;
}

//////////////////////////////////////////
//			Instance management			//
//////////////////////////////////////////

const int CART32M	= 32 * 1024 * 1024;
const int CART16M	= 16 * 1024 * 1024;
const int CART8M	= 8 * 1024 * 1024;
const int CART4M	= 4 * 1024 * 1024;

struct SInstance
{
	bool 	m_bTrimFF;
	bool 	m_bTrim00;
	bool 	m_bStopOnCloserRomSize;
	int		m_nStartOffset;
	string	m_szName;
	HPSF	m_hRomFile;
};

EXPORT_PSM HPSMI CreateInstance(HPSF hRomFile,HWND hParent)
{
	/*
	This function is called by pogoshell studio when it need to create a new instance of a modifier,
	each time you add a modifier to a file, pogoshell studio create an instance that must hold each own 
	data, usually you store those data in a custom struct

	hRomFile : an handle on the pogofile that will have the modifier apply on
	hParent : an handle on the windows that call CreateInstance, you can use it to display a dialog box,
			be careful if it's null that mean a dialog box is not welcome.

	*/
	SInstance * pInstance = new SInstance;
	pInstance->m_szName					= "Sample instance";
	pInstance->m_bTrim00				= true;
	pInstance->m_bTrimFF				= true;
	pInstance->m_bStopOnCloserRomSize	= true;
	pInstance->m_nStartOffset			= -1;
	pInstance->m_hRomFile				= hRomFile;
	return (HPSMI)pInstance;
}

EXPORT_PSM void DestroyInstance(HPSMI hInstance)
{
	/*
	This function is called by pogoshell studio when it need to delete an instance of a modifier,
	you should delete here the memory allocated by CreateInstance 
	*/

	SInstance * pInstance = (SInstance *)hInstance;
	delete pInstance;
}

EXPORT_PSM HPSMI LoadInstance(HPSF hRomFile,PSM_MEMORY_BLOCK MemBlock)
{
	/*
	This function is called by pogoshell studio when it need to Load an instance from a psw,
	hRomFile is passed here again cause it's not the same from one session to another, so you
	can't save it to the file.
	Please try to save/load your data wisely so that a new modifier can still open the old file

	hRomFile : an handle on the pogofile that will have the modifier apply on
	MemBlock : the MemoryBlock that contain the data to load
	*/

	SInstance * pInstance = NULL;
	MemoryBlock_Lock(MemBlock);
	PSMVERSION version = (PSMVERSION)MemoryBlock_ReadInt(MemBlock);
	if (version >= MakePsmVersion(0,22))
	{
		pInstance = new SInstance;
		pInstance->m_hRomFile				= hRomFile;
		pInstance->m_bTrimFF				= MemoryBlock_ReadBool(MemBlock);
		pInstance->m_bTrim00				= MemoryBlock_ReadBool(MemBlock);
		pInstance->m_bStopOnCloserRomSize	= MemoryBlock_ReadBool(MemBlock);
		pInstance->m_nStartOffset			= MemoryBlock_ReadInt(MemBlock);
		char szString[256];
		MemoryBlock_ReadString(MemBlock,szString,256);
		pInstance->m_szName					= szString;		
	}
	else
		pInstance = (SInstance *)CreateInstance(hRomFile,NULL);
	MemoryBlock_UnLock(MemBlock);

	return (HPSMI)pInstance;
}

EXPORT_PSM bool SaveInstance(HPSMI hInstance,PSM_MEMORY_BLOCK MemBlock)
{
	/*
	This function is called by pogoshell studio when it need to save an instance to a psw,
	Please don't save the handle to the pogofile because it's not the same from one session to another,
	Please try to save/load your data wisely so that a new modifier can still open the old file
	Don't forget to resize the MemoryBlock to the correct size at the end, by default it's 1024 bytes
	so if you don't you will save 1024 bytes of data to the file
	*/
	SInstance	* pInstance = (SInstance *)hInstance;

	MemoryBlock_Lock(MemBlock);
	MemoryBlock_WriteInt(MemBlock,(int)GetModifierVersion());
	MemoryBlock_WriteBool(MemBlock,pInstance->m_bTrimFF);
	MemoryBlock_WriteBool(MemBlock,pInstance->m_bTrim00);
	MemoryBlock_WriteBool(MemBlock,pInstance->m_bStopOnCloserRomSize);
	MemoryBlock_WriteInt(MemBlock,pInstance->m_nStartOffset);
	MemoryBlock_WriteString(MemBlock,pInstance->m_szName.c_str());
	size_t nSize = MemoryBlock_GetPos(MemBlock);
	MemoryBlock_UnLock(MemBlock);
	MemoryBlock_Resize(MemBlock,nSize);
	return true;
}

EXPORT_PSM const char * GetInstanceName(HPSMI hInstance)
{
	return ((SInstance *)hInstance)->m_szName.c_str();
}

EXPORT_PSM bool ConfigureInstance(HPSMI hInstance,HWND hParent)
{
	/*
	This function is called when the user click on configure on a modifier in the file properties, 
	if you don't implement this function, then the button will be disable.
	Usually this config dialog is use to set the option that are instance specific, like the name of
	the ips file to apply
	*/
	
	//DialogBoxParam(g_hDllInstance,MAKEINTRESOURCE(IDD_DIALOG_CONFIGURE_INSTANCE),hParent,(DLGPROC)ConfigureInstanceDlgProc,(LPARAM)hInstance);
	
	return true;
}

EXPORT_PSM bool ApplyModifier(HPSMI hInstance,PSM_MEMORY_BLOCK MemBlock)
{	
	/*This function is called pogoshell studio need to apply your modifier to a file, it might be more
	than one time. The calling order is defined by the order of the modifier in the file properties,
	don't hesitate to resize the memory block if you need to, and don't forgot to Unlock it once you
	are done
	
	Please respect the output convention that I use here, and don't use red for another things than error
	
	Output_Printf("Trimming \"%s\"... ",Psf_GetName(pInstance->m_hRomFile));
	Output_Printf("done\n");
	or
	Output_Printfc(false,RGB(255,0,0),"failed!\n");
	*/

	SInstance	* pInstance = (SInstance *)hInstance;

	Output_Printf("Trimming \"%s\"... ",Psf_GetName(pInstance->m_hRomFile));

	int nInitialSize = MemoryBlock_GetSize(MemBlock);
	int nCloserStandardCartSize = 0;
	if (pInstance->m_bStopOnCloserRomSize)
	{
		if (nInitialSize > CART32M)
			nCloserStandardCartSize = CART32M;
		else if (nInitialSize > CART16M)
			nCloserStandardCartSize = CART16M;
		else if (nInitialSize > CART8M)
			nCloserStandardCartSize = CART8M;
		else if (nInitialSize > CART4M)
			nCloserStandardCartSize = CART4M;
	}
	//if m_nStartOffset is equal to -1, then we need to find the start of the trash ourself
	int nStartTriming = pInstance->m_nStartOffset;
	if (nStartTriming < 0)
	{
		BYTE * pData = (BYTE *)MemoryBlock_Lock(MemBlock);

		int i = nInitialSize - 1;
		while (i > nCloserStandardCartSize)
		{
			if ( !((pInstance->m_bTrim00 && pData[i] == 0x00) || (pInstance->m_bTrimFF && pData[i] == 0xFF)))
				break;
			i--;
		}
		nStartTriming = i + 10; //+10 just in case the last few byte was necessary
		if (nStartTriming > nInitialSize)
			nStartTriming = nInitialSize;
		MemoryBlock_UnLock(MemBlock);
	}

	if (nStartTriming < nInitialSize)
		MemoryBlock_Resize(MemBlock,nStartTriming);

	Output_Printf("done\n");
	//Output_Printfc(false,RGB(255,0,0),"failed!\n");
	return true;
}


BOOL APIENTRY DllMain(HINSTANCE hInstance,DWORD  ul_reason_for_call,LPVOID lpReserved)
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		//store the dll instance to display dialog box later
		g_hDllInstance = hInstance;
	}
	return TRUE;
}